import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:circumstellar/src/connect_status.dart';
import 'package:circumstellar/src/discord_constants.dart';
import 'package:circumstellar/src/errors/resume_exception.dart';
import 'package:circumstellar/src/models/ws_payload.dart';
import 'package:circumstellar/src/rest/discord_rest.dart';
import 'package:circumstellar/src/ws/heartbeat.dart';
import 'package:w_transport/w_transport.dart' as transport;
import '../discord_client.dart';

/// A Discord WebSocket wrapper.
class DiscordWebSocketClient {
  /// The parent of this wrapper.
  DiscordClient _client;

  /// The REST client to accompany this wrapper.
  RestClient _rest;

  /// The internal WebSocket implementation.
  transport.WebSocket _ws;

  /// Defualt JSON decoder instance.
  JsonDecoder _decoder;

  /// Default JSON encoder instance.
  JsonEncoder _encoder;

  /// Handles automatically sending heartbeats using the heartbeat interval supplied.
  WSHeartbeater _hb;

  /// Tracks the current sequence number as sent from Discord.
  int currentSeq;

  /// Whether this client is attempting to resume.
  bool resuming = false;

  /// Whether this client has already identified.
  bool identified = false;

  /// A list of debug trace servers for the Hello payload.
  List<String> traceServers = [];

  /// The local cached gateway URL.
  String wsGatewayUri;

  /// Whether the gateway URL has been cached.
  bool get gatewayUriCached => wsGatewayUri != null;

  /// The current ID of the WebSocket session.
  String sessionId = null;

  /// Tracks whether Discord is currently replaying events to us.
  bool replayingEvents = false;

  /// Current status of the connection
  ConnectStatus status = ConnectStatus.notConnected;

  /// Default constructor.
  DiscordWebSocketClient(this._client, this._rest) {
    _decoder = new JsonDecoder();
    _encoder = new JsonEncoder();
    _createHandlers();
    _createCloseHandlers();
  }

  /// Terminates and closes the internal WebSocket and heartbeater.
  /// Will attempt to reconnect if [reconnect] is [true].
  Future stop([bool reconnect = false]) async {
    if (_ws != null) _ws.close(1000, 'User is disconnecting');
    _hb.stop();
    if (reconnect) {
      await connect(true);
    }
  }

  /// Attempts to initiate a WebSocket connection to Discord's gateway.
  Future connect([bool resuming = false]) async {
    this.resuming = resuming;
    if (!gatewayUriCached) {
      this.wsGatewayUri = (await getGatewayUri()).toString();
    }
    try {
      this.status = ConnectStatus.connecting;
      this._ws = await transport.WebSocket.connect(Uri.parse(this.wsGatewayUri),
          transportPlatform: _client.transportPlatform);
    } on transport.WebSocketException {}
    this._ws.listen((data) async {
      await _handleWebSocketMessage(data);
    }, onError: (error) async {
      await _handleWebSocketError(error);
    }, onDone: () async {
      await _handleWebSocketClosed();
    });
  }

  Future _handleWebSocketError(dynamic error) async {
    await connect(true);
  }

  Future _handleWebSocketClosed() async {
    await connect(true);
  }

  /// A map of all handlers, for Discord events.
  Map<int, Function(WSPayloadData payload)> handlers = {};

  /// A map of close codes to handlers, for when the WebSocket is closed.
  Map<int, Function(int closeCode, String closeReason)> wsCloseHandlers = {};

  /// Loads all WebSocket close event handlers into memory.
  void _createCloseHandlers() {
    wsCloseHandlers = {
      4008: (int closeCode, String closeReason) {
        // 4008 : Rate Limited (WS)
      }
    };
  }

  /// Loads all handlers into memory.
  void _createHandlers() {
    handlers = {
      /**
       * Several OP codes are missing because we don't receive them, only send
       */
      1: (payload) {
        // Heartbeat
      },
      7: (payload) {
        // Reconnect
      },
      9: (payload) {
        status = ConnectStatus.notConnected;
        print('[Circumstellar] Invalid Session');
        // Invalid Session
      },
      10: (payload) {
        // Hello
        status = ConnectStatus.connected;
        _initializeHeartbeats(payload);
        traceServers.addAll(payload.innerPayload['_trace']);
      },
      11: (payload) {
        // Heartbeat ACK
      }
    };
  }

  /// Handles an incoming WebSocket message.
  Future _handleWebSocketMessage(dynamic rawData) async {
    var payload = WSPayloadData.build(_decoder.convert(rawData));

    if (payload.opCode.friendlyName == "Dispatch") {
      await _client.dispatcher.handleEvent(payload.eventName,
          payload.innerPayload, payload.innerPayloadIsJsonObject);
    } else {
      if (handlers.containsKey(payload.opCode.code)) {
        handlers[payload.opCode.code](payload);
      } else {
        print(
            'warn: Payload handler for op code ${payload.opCode.code} has not been registered');
      }
    }

    if (resuming) {
      status = ConnectStatus.connecting;
      _sendResume();
    }

    if (!resuming && !identified) {
      _sendIdentify();
    }
  }

  /// Sends a RESUME payload to reopen a failed connection.
  void _sendResume() {
    if (!resuming)
      throw new ResumeException(
          'Cannot attempt resume if DiscordClient.resuming is false.');
    print('Resuming dead connection');
    var payload = new WSPayloadData();
    payload.opCode = DiscordGatewayStatic.opCodes[6];
    Map<String, dynamic> data = {
      'token': this._client.auth.token,
      'session_id': this.sessionId,
      'seq': this.currentSeq
    };
    payload.innerPayload = data;
    payload.innerPayloadIsJsonObject = true;
    replayingEvents = true;
    sendMessage(payload);
  }

  /// Sends an IDENTIFY payload to initiate a connection.
  void _sendIdentify() {
    print('Identifying with gateway');
    identified = true;
    var payload = new WSPayloadData();
    payload.opCode = DiscordGatewayStatic.opCodes[2];
    Map<String, dynamic> data = {
      'token': this._client.auth.token,
      'properties': {
        '\$os': Platform.operatingSystem,
        '\$browser': 'Circumstellar',
        '\$device': 'Circumstellar'
      },
      'compress': false,
      'large_threshold': 250
    };
    payload.innerPayload = data;
    sendMessage(payload);
  }

  /// Sends a payload object (after converting to JSON) over the WebSocket.
  void sendMessage(WSPayloadData message) {
    _ws.add(_encoder.convert(message.buildOutgoing()));
  }

  /// Tells the heartbeater to being beating.
  void _initializeHeartbeats(WSPayloadData payload) {
    _hb =
        new WSHeartbeater(_client, payload.innerPayload['heartbeat_interval']);
    _hb.begin();
  }

  /// Uses the REST API to get an available gateway URL.
  Future<Uri> getGatewayUri() async {
    var response = await _rest.executeRoute(ApiRoutes.gateway.getGateway());
    return buildGatewayUri(Uri.parse(response.json['url']));
  }

  /// Constructs a gateway URL using the current compatible version.
  Uri buildGatewayUri(Uri received) {
    var str = received.toString();
    return Uri.parse(
        ("${str}/?v=${DiscordGatewayStatic.gatewayVersion}&encoding=json")); // Don't hardcode encoding=json
  }
}
